﻿/*
** ADOBE SYSTEMS INCORPORATED
** Copyright 2012 Adobe Systems Incorporated
** All Rights Reserved.
**
** NOTICE:  Adobe permits you to use, modify, and distribute this file in accordance with the
** terms of the Adobe license agreement accompanying it.  If you have received this file from a
** source other than Adobe, then your use, modification, or distribution of it requires the prior
** written permission of Adobe.
*/
package com.adobe.flascc
{
  import flash.display.Bitmap
  import flash.display.BitmapData
  import flash.display.DisplayObjectContainer;
  import flash.display.Sprite;
  import flash.display.StageScaleMode;
  import flash.events.Event;
  import flash.events.KeyboardEvent;
  import flash.events.MouseEvent;
  import flash.geom.Rectangle;
  import flash.net.FileReference;
  import flash.net.FileReferenceList;
  import flash.text.TextField;
  import flash.utils.ByteArray;

  import com.adobe.flascc.vfs.ISpecialFile;
  import com.adobe.flascc.vfs.RootFSBackingStore;

  /**
  * A basic implementation of a console for FlasCC apps.
  * The PlayerKernel class delegates to this for things like read/write
  * so that console output can be displayed in a TextField on the Stage.
  */
  public class Console extends Sprite implements ISpecialFile
  {
    private var _tf:TextField
    private var bm:Bitmap
    private var enableConsole:Boolean = true;
    private var runningInWorker:Boolean = false
    private var frameCount:int = 0
    private var enginetickptr:int;
    private var inputContainer
    private var bmd:BitmapData
    private var bmr:Rectangle
    private var keybytes:ByteArray = new ByteArray()
    private var vbuffer:int, kp:int;
	private var romFile:FileReference = new FileReference();
	private var biosFiles:FileReferenceList = new FileReferenceList();
	private var loaded:int = 0;
	private var romInited:Boolean, biosInited:Boolean;
	private var buttons:Vector.<Sprite> = new Vector.<Sprite>();
    private const emptyArgs:Vector.<int> = new Vector.<int>();
    
    /**
    * To Support the preloader case you might want to have the Console
    * act as a child of some other DisplayObjectContainer.
    */
    public function Console(container:DisplayObjectContainer = null)
    {
      CModule.rootSprite = container ? container.root : this

      if(CModule.runningAsWorker()) {
        return;
      }
/*
      if(container) {
        container.addChild(this)
        init(null)
      } else {
        addEventListener(Event.ADDED_TO_STAGE, init)
      }
*/
      if(container) {
        container.addChild(this)
        onAdded(null)
      } else {
        addEventListener(Event.ADDED_TO_STAGE, onAdded)
      }
    }
	
	private function getButton(label:String):Sprite
	{
		var result:Sprite = new Sprite();
		var tf:TextField = new TextField();
		tf.autoSize = "left";
		tf.textColor = 0xFFFFFF;
		tf.text = label;
		
		result.graphics.beginFill(0x0000FF);
		result.graphics.drawRect(0,0, tf.width, tf.height);
		result.graphics.endFill();
		result.addChild(tf);
		
		return result;
	}
	
	private function onAdded(e:Event):void
	{
		var btnRom:Sprite = getButton("ROM");
		btnRom.x = btnRom.y = 5;
		btnRom.addEventListener(MouseEvent.CLICK, onClickRom);

		buttons.push(btnRom);
		addChild(btnRom);
		
		if (enableConsole)
		{
			_tf = new TextField();
			_tf.multiline = true;
			_tf.width = stage.stageWidth - 10;
			_tf.height = stage.stageHeight - btnRom.height - 15;
			_tf.x = 5;
			_tf.y = btnRom.height + 10;
			addChild(_tf);
		}
	}
	
	private function ready():void
	{
		var btnStart:Sprite = getButton("START");
		btnStart.x = buttons[0].x + buttons[0].width + 5;
		btnStart.y = buttons[0].y;
		btnStart.addEventListener(MouseEvent.CLICK, start);
		
		buttons.push(btnStart);
		addChild(btnStart);
	}
	
	private function onClickRom(e:MouseEvent):void
	{
		romFile.addEventListener(Event.SELECT, onSelectFile);
		romFile.addEventListener(Event.COMPLETE, onCompleteFile);
		romFile.browse();
	}
	
	private function onSelectFile(e:Event):void
	{
		romFile.load();
	}
	
	private function onCompleteFile(e:Event):void
	{
		romFile.removeEventListener(Event.SELECT, onSelectFile);
		romFile.removeEventListener(Event.COMPLETE, onCompleteFile);
		consoleWrite(romFile.name + " LOADED\n");
		ready();
	}

    /**
    * All of the real FlasCC init happens in this method
    * which is either run on startup or once the SWF has
    * been added to the stage.
    */
    protected function start(e:MouseEvent):void
    {
		enableConsole = false;
		_tf.visible = false;
		for each (var btn:Sprite in buttons)
		{
			removeChild(btn);
		}
		buttons.length = 0;
		
		inputContainer = new Sprite()
		addChildAt(inputContainer, 0)

		addEventListener(Event.ENTER_FRAME, enterFrame)
		stage.addEventListener(KeyboardEvent.KEY_DOWN, bufferKeyDown)
		stage.addEventListener(KeyboardEvent.KEY_UP, bufferKeyUp)
		stage.frameRate = 60
		stage.scaleMode = StageScaleMode.NO_SCALE
		graphics.lineStyle(1, 0xe0e0e0)
		graphics.drawRect(0, 0, 480, 320)
		bmd = new BitmapData(1024,768, false)
		bm = new Bitmap(bmd)
		bm.scrollRect = new Rectangle(272,224,480,320);
//		bm.scaleX = bm.scaleY = 2;
		bmr = new Rectangle(0,0,bmd.width, bmd.height)
		bmd.fillRect(bmd.rect, 0)
		inputContainer.addChild(bm)

		CModule.vfs.addBackingStore(new RootFSBackingStore(), null);
		CModule.vfs.addFile("/"+romFile.name, romFile.data);
		try
		{
			runningInWorker = CModule.canUseWorkers;
			CModule.vfs.console = this

			if(runningInWorker) {
				CModule.startBackground(this, new <String>["VisualBoyAdvance.swf", "--filter-2xsai", romFile.name])
			} else {
				CModule.startAsync(this, new <String>["VisualBoyAdvance.swf", "--filter-2xsai", romFile.name])
			}
		}
		catch(e:*)
		{
			consoleWrite(e.toString() + "\n" + e.getStackTrace().toString())
			throw e
		}

		vbuffer = CModule.getPublicSymbol("__avm2_vgl_argb_buffer");
		enginetickptr = CModule.getPublicSymbol("main_loop");
    }

    /**
    * The callback to call when FlasCC code calls the posix exit() function. Leave null to exit silently.
    * @private
    */
    public var exitHook:Function;

    /**
    * The PlayerKernel implementation will use this function to handle
    * C process exit requests
    */
    public function exit(code:int):Boolean
    {
      // default to unhandled
      return exitHook ? exitHook(code) : false;
    }

    /**
    * The PlayerKernel implementation will use this function to handle
    * C IO write requests to the file "/dev/tty" (e.g. output from
    * printf will pass through this function). See the ISpecialFile
    * documentation for more information about the arguments and return value.
    */
    public function write(fd:int, bufPtr:int, nbyte:int, errnoPtr:int):int
    {
      var str:String = CModule.readString(bufPtr, nbyte)
      consoleWrite(str)
      return nbyte
    }

    /**
    * libVGL expects to be able to read Keyboard input from
    * file descriptor zero using normal C IO.
    */
    public function read(fd:int, bufPtr:int, nbyte:int, errnoPtr:int):int
    {
      if(fd == 0 && nbyte == 1) {
        keybytes.position = kp++
        if(keybytes.bytesAvailable) {
          CModule.write8(bufPtr, keybytes.readUnsignedByte())
          return 1
        } else {
          keybytes.length = 0
          keybytes.position = 0
          kp = 0
        }
      }
      return 0
    }

    /**
    * The PlayerKernel implementation will use this function to handle
    * C fcntl requests to the file "/dev/tty" 
    * See the ISpecialFile documentation for more information about the
    * arguments and return value.
    */
    public function fcntl(fd:int, com:int, data:int, errnoPtr:int):int
    {
      return 0
    }

    /**
    * The PlayerKernel implementation will use this function to handle
    * C ioctl requests to the file "/dev/tty" 
    * See the ISpecialFile documentation for more information about the
    * arguments and return value.
    */
    public function ioctl(fd:int, com:int, data:int, errnoPtr:int):int
    {
      return CModule.callI(CModule.getPublicSymbol("vglttyioctl"), new <int>[fd, com, data, errnoPtr]);
    }

    /**
    * Helper function that traces to the flashlog text file and also
    * displays output in the on-screen textfield console.
    */
    protected function consoleWrite(s:String):void
    {
      trace(s)
      if(enableConsole) {
        _tf.appendText(s)
        _tf.scrollV = _tf.maxScrollV
      }
    }

    public function bufferKeyDown(ke:KeyboardEvent) {
      keybytes.writeByte(int(ke.keyCode & 0x7F))
    }
    
    public function bufferKeyUp(ke:KeyboardEvent) {
      keybytes.writeByte(int(ke.keyCode | 0x80))
    }
    
    protected function enterFrame(e:Event):void
    {
      if(runningInWorker) {
        CModule.serviceUIRequests()
        if(vbuffer == 0)
          vbuffer = CModule.getPublicSymbol("__avm2_vgl_argb_buffer")
      } else {
        CModule.callI(enginetickptr, emptyArgs)
      }

      var ram:ByteArray = CModule.ram;
      ram.position = CModule.read32(vbuffer)
      if (ram.position != 0) {
        frameCount++
        bmd.setPixels(bmr, ram)
      }
    }

    /**
    * Provide a way to get the TextField's text.
    */
    public function get consoleText():String
    {
        var txt:String = null;

        if(_tf != null){
            txt = _tf.text;
        }
        
        return txt;
    }
  }
}
